/*	SPKMeans.cc 

	Implementation of the SPKMeans class
*/

#include <time.h>
#include <assert.h>
#include <math.h>
#include <iostream.h>
#include <stdio.h>
#include <fstream.h>
#include <string.h>
#include "mat_vec.h"
#include "SPKMeans.h"

extern long memory_consume;

//#define CV_Sim(i, j) CV_Sim[(i>=j)?(i*(i+1)/2+j):(j*(j+1)/2+i)]

SPKMeans::SPKMeans(SparseMatrixDouble *p_Docs, int cluster[], int num_clusters,
		int init_method, float epsilon)
{
	assert(p_Docs != NULL && cluster != NULL && num_clusters > 0 &&  epsilon > 0);

	//pDocs = p_docs;
	Cluster = cluster;
	n_Clusters = num_clusters;
	Init_Method = init_method;
	Epsilon = epsilon;

	n_Words = p_Docs->GetNumRow();
	n_Docs = p_Docs->GetNumCol();
	n_Iters = 0;
	Result = 0;
	Alg = SPHERICAL_K_MEANS;
	Perturb = DEFAULT_PERTURB;
	n_Samples = DEFAULT_NUM_SAMPLE;
	SampleSize = DEFAULT_SAMPLE_SIZE;
	EncodeScheme = NORM_TERM_FREQ;
	ObjFunc = NON_WEIGHTED_OBJ;
	DumpCSize = false;

	Inner_Sim = new float[n_Docs];
	//CV_Sim = new float[n_Clusters * (n_Clusters+1) / 2];
	Sim_Mat = new (float *)[n_Clusters];
	for (int j = 0; j < n_Clusters; j++)
		Sim_Mat[j] = new float[n_Docs];
	memory_consume+=(n_Docs+n_Clusters*n_Docs/2)*sizeof(float);

	if (init_method != CONCEPT_VECTORS_INIT)
	{
		Concept_Vector = new VECTOR_double[n_Clusters];
		for (int i = 0; i < n_Clusters; i++)
			Concept_Vector[i] = new float[n_Words];
	}

	qTransA_Vector = cTransA_Vector = NULL;

	memory_consume+=n_Clusters*n_Words*sizeof(float)*2;

	old_CV = new VECTOR_double[n_Clusters];
	for (int i = 0; i < n_Clusters; i++)
		old_CV[i] = new float[n_Words];
	diff = new float[n_Clusters];
	ClusterSize = new  int[n_Clusters];
	DocNormSqr = new float[n_Docs];
	CVNormSqr = new float[n_Clusters];

	memory_consume+=(n_Clusters*2+n_Docs+n_Clusters)*sizeof(float);
	//rand_gen.Set((unsigned)0);

	rand_gen.Set((unsigned)time(NULL));
}

SPKMeans::SPKMeans(DenseMatrixDouble *Dp_Docs, int cluster[], int num_clusters,
		int init_method, float epsilon)
{
  assert(Dp_Docs != NULL && cluster != NULL && num_clusters > 0 &&  epsilon > 0);

  //DpDocs = p_docs;
	Cluster = cluster;
	n_Clusters = num_clusters;
	Init_Method = init_method;
	Epsilon = epsilon;

	n_Words = Dp_Docs->GetNumRow();
	n_Docs = Dp_Docs->GetNumCol();
	n_Iters = 0;
	Result = 0;

	Inner_Sim = new float[n_Docs];
	//CV_Sim = new float[n_Clusters * (n_Clusters+1) / 2];
	Sim_Mat = new (float *)[n_Clusters];
	for (int j = 0; j < n_Clusters; j++)
		Sim_Mat[j] = new float[n_Docs];

	if (init_method != CONCEPT_VECTORS_INIT)
	{
		Concept_Vector = new VECTOR_double[n_Clusters];
		for (int i = 0; i < n_Clusters; i++)
			Concept_Vector[i] = new float[n_Words];
	}

	qTransA_Vector = cTransA_Vector =NULL;

	old_CV = new VECTOR_double[n_Clusters];
	for (int i = 0; i < n_Clusters; i++)
		old_CV[i] = new float[n_Words];
	diff = new float[n_Clusters];

	ClusterSize = new  int[n_Clusters];
	DocNormSqr = new float[n_Docs];
	CVNormSqr = new float[n_Clusters];

	rand_gen.Set((unsigned)time(NULL));
}

SPKMeans::SPKMeans()
{
	Alg = SPHERICAL_K_MEANS;
	n_Samples = DEFAULT_NUM_SAMPLE;
	Perturb = DEFAULT_PERTURB;
	SampleSize = DEFAULT_SAMPLE_SIZE;
	Init_Method = RANDOM_PERTURB_INIT;
	EncodeScheme = NORM_TERM_FREQ;
	ObjFunc = NON_WEIGHTED_OBJ;
	DumpCSize = false;
	Concept_Vector = NULL;

	Inner_Sim = NULL;
	//CV_Sim = NULL;

	rand_gen.Set((unsigned)time(NULL));
}

void SPKMeans::Init(SparseMatrixDouble *p_Docs, int cluster[], int num_clusters,
		int init_method, float epsilon)
{
	assert(p_Docs != NULL && cluster != NULL && num_clusters > 0 &&  epsilon > 0);

	if (Concept_Vector != NULL && Init_Method != CONCEPT_VECTORS_INIT)
	{
	  for (int i = 0; i < n_Clusters; i++)
	    delete[] Concept_Vector[i];
	  delete[] Concept_Vector;
	}
	if (Sim_Mat!= NULL)
	{
	  for (int i = 0; i < n_Clusters; i++)
	    delete[] Sim_Mat[i];
	  delete[] Sim_Mat;
	}
	
	if (qTransA_Vector != NULL)
	  {
	    for (int i = 0; i < n_Clusters; i++)
	      delete[] qTransA_Vector[i];
	    delete[] qTransA_Vector;
	  }

	
	//pDocs = p_docs;
	Cluster = cluster;
	n_Clusters = num_clusters;
	Init_Method = init_method;
	Epsilon = epsilon;

	n_Words = p_Docs->GetNumRow();
	n_Docs = p_Docs->GetNumCol();
	n_Iters = 0;
	Result = 0;

	Inner_Sim = new float[n_Docs];
	//CV_Sim = new float[n_Clusters * (n_Clusters+1) / 2];
	Sim_Mat = new (float *)[n_Clusters];
	for (int j = 0; j < n_Clusters; j++)
		Sim_Mat[j] = new float[n_Docs];

	if (init_method != CONCEPT_VECTORS_INIT)
	{
		Concept_Vector = new VECTOR_double[n_Clusters];
		for (int i = 0; i < n_Clusters; i++)
			Concept_Vector[i] = new float[n_Words];
	}

	qTransA_Vector = cTransA_Vector =NULL;
	old_CV = new VECTOR_double[n_Clusters];
	for (int i = 0; i < n_Clusters; i++)
		old_CV[i] = new float[n_Words];
	diff = new float[n_Clusters];

	ClusterSize = new  int[n_Clusters];
	DocNormSqr = new float[n_Docs];
	CVNormSqr = new float[n_Clusters];

	rand_gen.Set((unsigned)time(NULL));
}

void SPKMeans::Init(DenseMatrixDouble *Dp_Docs, int cluster[], int num_clusters,
		int init_method, float epsilon)
{
  assert(Dp_Docs != NULL && cluster != NULL && num_clusters > 0 &&  epsilon > 0);
  //cout<<"fine";
  if (Concept_Vector != NULL && Init_Method != CONCEPT_VECTORS_INIT)
    {
      for (int i = 0; i < n_Clusters; i++)
	{
	  delete[] Concept_Vector[i];
	 
	}
      
      delete[] Concept_Vector;
      
    }
  //cout<<"fine";
  
  if (Sim_Mat!= NULL)
	{
	  for (int i = 0; i < n_Clusters; i++)
	    delete[] Sim_Mat[i];
	  delete[] Sim_Mat;
	  }
  /*if (qTransA_Vector!= NULL)
    {
      for (int i = 0; i < n_Clusters; i++)
      	delete[] qTransA_Vector[i];
      delete[] qTransA_Vector;
      }*/
  //cout<<"fine";
	//DpDocs = p_docs;
	Cluster = cluster;
	n_Clusters = num_clusters;
	Init_Method = init_method;
	Epsilon = epsilon;

	n_Words = Dp_Docs->GetNumRow();
	n_Docs = Dp_Docs->GetNumCol();

	//cout<<"\n"<< Dp_Docs->GetNumRow()<<" "<<Dp_Docs->val(0,0)<<endl;
	n_Iters = 0;
	Result = 0;

	Inner_Sim = new float[n_Docs];
	//CV_Sim = new float[n_Clusters * (n_Clusters+1) / 2];
	Sim_Mat = new (float *)[n_Clusters];
	for (int j = 0; j < n_Clusters; j++)
		Sim_Mat[j] = new float[n_Docs];

	if (init_method != CONCEPT_VECTORS_INIT)
	{
		Concept_Vector = new VECTOR_double[n_Clusters];
		for (int i = 0; i < n_Clusters; i++)
			Concept_Vector[i] = new float[n_Words];
	}

	qTransA_Vector = cTransA_Vector =NULL;

	old_CV = new VECTOR_double[n_Clusters];
	for (int i = 0; i < n_Clusters; i++)
		old_CV[i] = new float[n_Words];
	diff = new float[n_Clusters];

	ClusterSize = new  int[n_Clusters];
	DocNormSqr = new float[n_Docs];
	CVNormSqr = new float[n_Clusters];

	rand_gen.Set((unsigned)time(NULL));
}

SPKMeans::~SPKMeans()
{
  if (Concept_Vector != NULL && Init_Method != CONCEPT_VECTORS_INIT)
    {
      for (int i = 0; i < n_Clusters; i++)
	{
	  delete[] Concept_Vector[i];
	  delete[] Sim_Mat[i];
	}
      
      delete[] Concept_Vector;
      delete[] Sim_Mat;
    }

  if (qTransA_Vector!= NULL)
    {
      for (int i = 0; i < n_Clusters; i++)
      	delete[] qTransA_Vector[i];
      delete[] qTransA_Vector;
    }
  if (cTransA_Vector!= NULL)
    {
      for (int i = 0; i < n_Clusters; i++)
      	delete[] cTransA_Vector[i];
      delete[] cTransA_Vector;
    }

  
  delete[] ClusterSize; // any problem? ...
  delete[] DocNormSqr;
  delete[] CVNormSqr;
  delete[] Inner_Sim;
	//delete[] CV_Sim;
}


float ** SPKMeans::getQTA()
{
  return qTransA_Vector;
}

float ** SPKMeans::getCTA()
{
  return cTransA_Vector;
}

void SPKMeans::checkCTA(SparseMatrixDouble *p_Docs)
{
  if (cTransA_Vector == NULL)
    {
      int i;

      cout<<"\nDoing Concept Decomposition without QR ..."<<endl;

      cTransA_Vector = new VECTOR_double[n_Clusters];
      for (i = 0; i < n_Clusters; i++)
	cTransA_Vector[i] = new float[n_Docs];
      for( i=0; i<n_Clusters; i++)
	p_Docs->trans_mult(Concept_Vector[i], cTransA_Vector[i]);
     
      cout<<"Finished Concept Decomposition.\n"<<endl;
    }
}

void SPKMeans::checkQTA(SparseMatrixDouble *p_Docs)
{
  if (qTransA_Vector == NULL)
    {
      int i, j;
      //float **cTransA_Vector;
      float **C_Vector, **Q_Vector, **QTrans_Vector;
      float *work;

      cout<<"\nDoing Concept Decomposition with QR ..."<<endl;

      qTransA_Vector = new VECTOR_double[n_Clusters];
      for (i = 0; i < n_Clusters; i++)
	qTransA_Vector[i] = new float[n_Docs];
      
      C_Vector = new VECTOR_double[n_Words];
      Q_Vector = new VECTOR_double[n_Words];
      for (i = 0; i < n_Words; i++)
	{
	  C_Vector[i] = new float[n_Clusters];
	  Q_Vector[i] = new float[n_Clusters];
	}
      
      for ( i=0; i<n_Clusters; i++)
	for ( j=0; j<n_Words; j++)
	  C_Vector[j][i]=Concept_Vector[i][j];
      
      work = new float[n_Words];
      dqrbasis( n_Words, n_Clusters, C_Vector, Q_Vector , work);
      
      QTrans_Vector = new VECTOR_double[n_Clusters];
      for (i = 0; i < n_Clusters; i++)
	QTrans_Vector[i] = new float[n_Words];
      
      for ( i=0; i<n_Clusters; i++)
	for ( j=0; j<n_Words; j++)
	  QTrans_Vector[i][j]=Q_Vector[j][i];
      
      
      // Q^T*A matrix is K * D
      
      for( i=0; i<n_Clusters; i++)
	p_Docs->trans_mult(QTrans_Vector[i], qTransA_Vector[i]);
   
      cout<<"Finished Concept Decomposition.\n"<<endl;
    }
}

void SPKMeans::Concept_Dec(SparseMatrixDouble *p_Docs,  char * filenamePrefix, int meth)
{
  char matrix_file[256];
  int i, j;
  //float **cTransA_Vector;
  float **C_Vector, **Q_Vector, **QTrans_Vector;
  float *work;


  qTransA_Vector = new VECTOR_double[n_Clusters];
  for (i = 0; i < n_Clusters; i++)
    qTransA_Vector[i] = new float[n_Docs];

  cTransA_Vector = new VECTOR_double[n_Clusters];
  for (i = 0; i < n_Clusters; i++)
    cTransA_Vector[i] = new float[n_Docs];

  if (meth == CD_QR || meth ==CON_DEC)
    {    
      cout<<"Generating C^T matrix..."<<endl;
      sprintf(matrix_file,"%s%s%d", filenamePrefix, "_cTrans.", n_Clusters);
      std::ofstream cTrans(matrix_file);
      // C^T is K * W matrix
      for ( i=0; i<n_Clusters; i++)
	{
	  for ( j=0; j<n_Words; j++)
	    cTrans<<Concept_Vector[i][j]<<"\t";
	  cTrans<<endl;
	}
      
      cout<<"Generating C^T*A matrix..."<<endl;
      // C^T*A is K * D matrix
      
      for( i=0; i<n_Clusters; i++)
	p_Docs->trans_mult(Concept_Vector[i], cTransA_Vector[i]);

      sprintf(matrix_file,"%s%s%d", filenamePrefix, "_cTransA.", n_Clusters);
      std::ofstream cTransA(matrix_file);
      for ( i=0; i<n_Clusters; i++)
	{
	  for ( j=0; j<n_Docs; j++)
	    cTransA<<cTransA_Vector[i][j]<<"\t";
	  cTransA<<endl;
	}

       cout<<"Generating A^T*C matrix..."<<endl;
       // C^T*A is D * K matrix

       sprintf(matrix_file,"%s%s%d", filenamePrefix, "_aTransC.", n_Clusters);
       std::ofstream aTransC(matrix_file);
       for ( i=0; i<n_Docs; i++)
	 {
	   for ( j=0; j<n_Clusters; j++)
	     aTransC<<cTransA_Vector[j][i]<<"\t";
	   aTransC<<endl;
	 }
    }
   if (meth == QR_DEC || meth == CD_QR)
    {
      cout<<"Generating Q^T matrix..."<<endl;
      // Q^T matrix is W * K
      
      C_Vector = new VECTOR_double[n_Words];
      Q_Vector = new VECTOR_double[n_Words];
      for (i = 0; i < n_Words; i++)
	{
	  C_Vector[i] = new float[n_Clusters];
	  Q_Vector[i] = new float[n_Clusters];
	}
      
      for ( i=0; i<n_Clusters; i++)
	for ( j=0; j<n_Words; j++)
	  C_Vector[j][i]=Concept_Vector[i][j];
      
      work = new float[n_Words];
      dqrbasis( n_Words, n_Clusters, C_Vector, Q_Vector , work);
      
      sprintf(matrix_file,"%s%s%d", filenamePrefix, "_qTrans.", n_Clusters);
      std::ofstream qTrans(matrix_file);
      for ( i=0; i<n_Clusters; i++)
	{
	  for ( j=0; j<n_Words; j++)
	    qTrans<<Q_Vector[j][i]<<"\t";
	  qTrans<<endl;
	}
      
      QTrans_Vector = new VECTOR_double[n_Clusters];
      for (i = 0; i < n_Clusters; i++)
	QTrans_Vector[i] = new float[n_Words];
      
      for ( i=0; i<n_Clusters; i++)
	for ( j=0; j<n_Words; j++)
	  QTrans_Vector[i][j]=Q_Vector[j][i];
      
      cout<<"Generating Q^T*A matrix..."<<endl;
      // Q^T*A matrix is K * D
      
      for( i=0; i<n_Clusters; i++)
	p_Docs->trans_mult(QTrans_Vector[i], qTransA_Vector[i]);
      
      sprintf(matrix_file,"%s%s%d", filenamePrefix, "_qTransA.", n_Clusters);
      std::ofstream qTransA(matrix_file);
      for ( i=0; i<n_Clusters; i++)
	{
	  for ( j=0; j<n_Docs; j++)
	    qTransA<<qTransA_Vector[i][j]<<"\t";
	  qTransA<<endl;
	}
  
    }
}

void SPKMeans::dump()
{
	int i;

	for (i = 0; i < n_Clusters; i++)
		cout << "Cluster " << i << ": " << ClusterSize[i] << "\n";
//	for (i = 0; i < n_Docs; i++)
//		cout << Cluster[i] << " ";
//	cout << "\n";
	cout << "objective function value :" << GetResult() << "\n";
}



int SPKMeans::remove_Empty_Clusters(SparseMatrixDouble *p_Docs)
{
  int i, j, k, n_Empty_Clusters, Empty_Clusters_ID[n_Clusters];
  
  for (i=0, n_Empty_Clusters=0; i<n_Clusters; i++)
    {
      
      if (ClusterSize[i] == 0)
	{	
	  Empty_Clusters_ID[n_Empty_Clusters] = i;
	  
	  n_Empty_Clusters++;
	}
    }
  if(n_Empty_Clusters > 0)
    if(n_Docs >= n_Clusters)
      {
	int *rank = new int [n_Docs] ;
	
	//sort the dot products; output in is index
	ranksa_c(Inner_Sim, rank, &n_Docs);
			
	//rearrange the cluster ID for doc in non-empty clusters by shifting cluster ID to the 0 side
	
	for(j =0; j<n_Docs && Inner_Sim[rank[j]]==0.0; j++)
	  
	
	for (i = 0; i < n_Empty_Clusters; i++)
	  Cluster[rank[i+j]] = Empty_Clusters_ID[i];
	 
	// reset concept vectors
	for (i = 0; i < n_Clusters; i++)
	  for (j = 0; j < n_Words; j++)
	    Concept_Vector[i][j] = 0.0;

	k=0;

	// compute new concept vectors
	for (i = 0; i < n_Docs; i++)
	  {
	    while (i<empty_Docs_ID[k])
	      {
		assert(Cluster[i] >=0 && Cluster[i] < n_Clusters);
		for (j = p_Docs->col_ptr(i); j < p_Docs->col_ptr(i+1); j++)
		  Concept_Vector[Cluster[i]][p_Docs->row_ind(j)]+= p_Docs->val(j);
		i++;
	      }
	    k++;
	  }
			
	for (i = 0; i < n_Clusters; i++)
	  normalize_vec(Concept_Vector[i], n_Words);

	for (i = 0; i < n_Clusters; i++)
	  diff[i] = 0.0;
	
	for (i = 0; i < n_Docs; i++)
	  Inner_Sim[i] = dot_mult(p_Docs, i, Concept_Vector[Cluster[i]]);
				
			
	for (i = 0; i < n_Clusters; i++)
	  p_Docs->trans_mult(Concept_Vector[i], Sim_Mat[i]);	
	
	cout<<"Fixed "<<n_Empty_Clusters<< " empty clusters..."<<endl;
      }
    else
      cout<<"More clusters("<<n_Clusters<<") than documents("<<n_Docs<<"). Empty clusters not fixed."<<endl;
  
  return n_Empty_Clusters;
}

void SPKMeans::generate_Confusion_Matrix (int *label, int n_Class)
{
  int **n_l, i, j, k;
  
   // generate the confusion matrix
    if(n_Class >0)
      {

	cout<<endl<<endl<<"The confusion matrix looks:"<<endl;
	n_l=new (int *)[n_Clusters];
	for(i=0; i<n_Clusters;i++)
	  {  
	    n_l[i]=new int[n_Class];
	    for (j=0; j<n_Class; j++)
	      n_l[i][j]=0;
	  }
	k=0;
	for(i=0; i<n_Docs; i++)
	  {
	    while (i<empty_Docs_ID[k])
	    {
	      n_l[Cluster[i]][label[i]]++;
	      i++;
	    }
	    k++;
	  }
	for (i = 0; i < n_Clusters; i++)
	  {
	    printf("\n#%d", i);
	    for (j = 0; j < n_Class; j++)
	      printf("\t%d", n_l[i][j]);
	    fprintf(stdout, "\n");
	  }
	cout<<endl<<endl;
	/******************************/
	int max, cumu_max, cumu_nl;
	double sum, mut_info=0.0;
	for(i=0;i<n_Clusters; i++)
	  {
	    sum=0.0;
	    max=-1;
	    for(j=0;j<n_Class;j++)
	      {
		if (max<n_l[i][j])
		  max=n_l[i][j];
		if(ClusterSize[i]!=0 && n_l[i][j]!=0)
		  sum+=(double)n_l[i][j]/ClusterSize[i]*log((double)n_l[i][j]/ClusterSize[i])/log((double)n_Class);
	      }
	    if(ClusterSize[i]!=0)
	      {  
		cout<<"Purity of cluster "<<i<<" is :"<<(double)max/ClusterSize[i]<<endl;
		cout<<"Entropy of cluster "<<i<<" is :"<<-1*sum<<endl;
	      } 
	  }
	for(i=0; i<n_Clusters; i++)
	  {
      
	    for(int k=0;k<n_Class;k++)
	      { 
		cumu_max=cumu_nl=0;
		
		for (j=0; j< n_Clusters; j++)
		  {
		    cumu_max+=n_l[j][k];
		  }
		for(j=0;j<n_Class;j++)
		  {
		    cumu_nl+=n_l[i][j];
		  }
		if(n_l[i][k]!=0)
		  mut_info+=n_l[i][k]*log((double)n_Docs*n_l[i][k]/cumu_max/cumu_nl)/log((double)n_Clusters*n_Class);
	      }
	  }
	cout<<"Mutual info. of this clustering is : "<<mut_info/n_Docs<<endl;


	for(i=0; i<n_Clusters;i++)
	  {
	    delete [] n_l[i];
	  }
	delete [] n_l;
      }
    cout<<endl<<endl;
}

int SPKMeans::read_cate(char *fname, int * label)
{
  int i, max=0, n_Doc;
  char whole_line[256];

  std::ifstream catefile(fname);

  if(catefile==0)
    {
      cout<<"Category file "<<fname<<" can't open.\n"<<endl;
      return -1;
      //exit(1);
    }
  catefile>>n_Doc;
  catefile.getline(whole_line, 256, '\n');
  //catefile>>i;

  if(n_Doc != n_Docs)
    {
      cout<<"Wrong number of docs in true label file!!!"<<endl;
      return -1;
    }

  cout<<"True label file: "<<fname<<endl;

  for(i=0; i<n_Doc; i++)
  {
    catefile>>label[i];
    catefile.getline(whole_line, 256, '\n');
    if(label[i]>max)
      max=label[i];
  }

  catefile.close();
  return max;
}


void SPKMeans::Iterate(SparseMatrixDouble *p_Docs, char *initial_seed, char *true_label)
{
  float temp_result;
  int i, j, k;
  clock_t start_clock, finish_clock;
  int *label= new int[n_Docs], n_Class=0;

  std::ifstream gpfile(initial_seed);
  n_Class = read_cate(true_label, label) +1;

  // normalize document matrix if using spherical k means algorithm
  // else compute norms of document vectors
  if (Alg == SPHERICAL_K_MEANS)
    normalize_mat(p_Docs);
  else
    {
      for (i = 0; i < n_Docs; i++)
	{
	  DocNormSqr[i] = 0;
	  for (j = p_Docs->col_ptr(i); j < p_Docs->col_ptr(i+1); j++)
	    DocNormSqr[i] += p_Docs->val(j) * p_Docs->val(j);
	}
    }

  switch (Init_Method)
    {
    case SAMPLE_INIT:
      random_perturb_init(p_Docs);
		//refined_partition(p_Docs,n_Samples);
      break;
    case RANDOM_PERTURB_INIT:
      random_perturb_init(p_Docs);
      for(i=0; i<n_Docs; i++)
	Cluster[i]=0;
      break;
    case RANDOM_INIT:
      random_init(p_Docs);
      if (Alg == EUCLIDEAN_K_MEANS)	// should be moved into the concept vector initialization
	{
	  for (i = 0; i < n_Clusters; i++)
	    {
	      CVNormSqr[i] = 0;
	      for (j = 0; j < n_Words; j++)
		CVNormSqr[i] += Concept_Vector[i][j] * Concept_Vector[i][j];
	    }
	}
      break;
    case FILE_INIT:
      if(gpfile.is_open())
	{
	  cout<<"Initial cluster ID file: "<<initial_seed<<endl;
	  gpfile>>i>>j;
	  
	  if(i!=n_Docs)
	    {
	      cout<<"Wrong number of document!!! So using random initialization ..."<<endl;
	      random_perturb_init(p_Docs);
	      for (i = 0; i < n_Docs; i++)
		{
		  Cluster[i] = 0;
		}
	    }
	  else{
	
	    char whole_line[256];
	    
	    for (i=0; i<n_Docs; i++)
	      {
		gpfile>>Cluster[i];
		gpfile.getline(whole_line, 256, '\n');
		
	      }
	    
	    gpfile.close();

	    // reset concept vectors
	    for (i = 0; i < n_Clusters; i++)
	      for (j = 0; j < n_Words; j++)
		Concept_Vector[i][j] = 0.0;

	    k=0;
	    // compute new concept vectors
	    for (i = 0; i < n_Docs; i++)
	      {
		while (i<empty_Docs_ID[k])
		  {
		    if((Cluster[i] >=0) && (Cluster[i] < n_Clusters))
		      {
			assert(Cluster[i] >=0 && Cluster[i] < n_Clusters);
			for (j = p_Docs->col_ptr(i); j < p_Docs->col_ptr(i+1); j++)
			  {
			    Concept_Vector[Cluster[i]][p_Docs->row_ind(j)]+= p_Docs->val(j);
			  }
		      }
		    else
		      Cluster[i] =0;
		  
		    i++;
		  }
		k++;
	      }
	  
	  }

       
	  if (Alg == SPHERICAL_K_MEANS)
	    for (i = 0; i < n_Clusters; i++)
	      normalize_vec(Concept_Vector[i], n_Words);

	  compute_cluster_size();
	  outputClusterSize();
	  Result = coherence(p_Docs);
	  cout<<"Initial Obj. func.: "<<Result<<endl;
	  cout<<"Initial confusion matrix :"<<endl;
	  generate_Confusion_Matrix (label, n_Class);
	}
      break;
    default:
      break;
    }

  int n_Empty_Clusters=0;
  // initial assignment, all documents are assigned to cluster 0
  for (i = 0; i < n_Docs; i++)
    {
      Inner_Sim[i] = dot_mult(p_Docs, i, Concept_Vector[Cluster[i]]);
      
    }
 
  spkmeans_worked=true;
  for (i = 0; i < n_Clusters; i++)
    p_Docs->trans_mult(Concept_Vector[i], Sim_Mat[i]);
  for (i = 0; i < n_Clusters; i++)
    diff[i] = 0.0;

	/* iterate until the stopping criterion is met
	   the stopping criterion adopted here is to stop
	   if the "change" in objective function after an iteration
	   of the algorithm is less than a certain threshold. */
	Result = 0.0;
	n_Iters = 0;
//	pDocs->prepare_trans_mult();
	start_clock = clock();
	do
	{
	  assign_cluster(p_Docs);
	  compute_cluster_size();
	  n_Empty_Clusters= remove_Empty_Clusters(p_Docs);
	  
	  temp_result = Result;
	  n_Iters++;

	  
	  // save old concept vectors
	  for (i = 0; i < n_Clusters; i++)
	    for (j = 0; j < n_Words; j++)
	      old_CV[i][j] = Concept_Vector[i][j];
	  if (n_Empty_Clusters ==0)
	    {
	      // reset concept vectors
	      for (i = 0; i < n_Clusters; i++)
		for (j = 0; j < n_Words; j++)
		  Concept_Vector[i][j] = 0.0;
	      
	      k=0;
	      // compute new concept vectors
	      for (i = 0; i < n_Docs; i++)
		{
		  while (i<empty_Docs_ID[k])
		    {
		      assert(Cluster[i] >=0 && Cluster[i] < n_Clusters);
		      for (j = p_Docs->col_ptr(i); j < p_Docs->col_ptr(i+1); j++)
			Concept_Vector[Cluster[i]][p_Docs->row_ind(j)]+= p_Docs->val(j);
		      i++;
		    }
		  k++;
		}
	      
	      if (Alg == SPHERICAL_K_MEANS)
		for (i = 0; i < n_Clusters; i++)
		  normalize_vec(Concept_Vector[i], n_Words);
	      else
		{
		  for (i = 0; i < n_Clusters; i++)
		    {
		      CVNormSqr[i] = 0;
		  if (ClusterSize[i] != 0)
		    for (j = 0; j < n_Words; j++)
		      {
			Concept_Vector[i][j] /= ClusterSize[i];
			CVNormSqr[i] += Concept_Vector[i][j]
			  * Concept_Vector[i][j];
		      }
		    }
		}
	    
	      // compute the difference between old and new concept vectors
	      for (i = 0; i < n_Clusters; i++)
		{
		  diff[i] = 0.0;
		  for (j = 0; j < n_Words; j++)
		    diff[i] += (old_CV[i][j] - Concept_Vector[i][j]) * (old_CV[i][j] - Concept_Vector[i][j]);
		  diff[i] = sqrt(diff[i]);
		}
	    }

	  // compute new objective function
	  Result = coherence(p_Docs);
	  if(Dump)
	    cout<<"Objective function: "<<Result<<endl;
	  if (DumpCSize) dump();
	  
	} while (//(temp_result > Result || Alg != EUCLIDEAN_K_MEANS) &&
		(fabs(temp_result - Result) > Epsilon*Result));
	finish_clock = clock();

	if(Dump)
	  generate_Confusion_Matrix (label, n_Class);
	
	if (n_Iters<=2)
	  spkmeans_worked =false;

	cout<<"Core kmeans (routine) time: "<<(finish_clock - start_clock)/1e6<<endl;
	float *word_den = new float[n_Clusters];
	for (i = 0; i < n_Clusters; i++)
	{
		word_den[i] = 0.0;
		for (j = 0; j < n_Words; j++)
			if (Concept_Vector[i][j] != 0) word_den[i] += 1.0/n_Words;
	}

	cout << "word density of concept vectors: ";
	for (i = 0; i < n_Clusters; i++)
		cout << word_den[i] << " ";
	cout << "\n";
}


void SPKMeans::Iterate(SparseMatrixDouble *p_Docs, char * filename)
{
  float temp_result;
  int i, j;
  clock_t start_clock, finish_clock;
  
  // normalize document matrix if using spherical k means algorithm
  // else compute norms of document vectors
  if (Alg == SPHERICAL_K_MEANS)
    normalize_mat(p_Docs);
  else
    {
      for (i = 0; i < n_Docs; i++)
	{
	  DocNormSqr[i] = 0;
	  for (j = p_Docs->col_ptr(i); j < p_Docs->col_ptr(i+1); j++)
	    DocNormSqr[i] += p_Docs->val(j) * p_Docs->val(j);
	}
    }

  std::ifstream gpfile(filename);
  gpfile>>i>>j;

  if(i!=n_Docs)
  {
    cout<<"Wrong number of document!!!"<<endl;
    exit(1);
  }
  
  for (i=0; i<n_Docs; i++)
    gpfile>>Cluster[i];

  // reset concept vectors
  for (i = 0; i < n_Clusters; i++)
    for (j = 0; j < n_Words; j++)
      Concept_Vector[i][j] = 0.0;

  // compute new concept vectors
  for (i = 0; i < n_Docs; i++)
    {
      //assert(Cluster[i] >=0 && Cluster[i] < n_Clusters);
      if(Cluster[i] >=0)
	for (j = p_Docs->col_ptr(i); j < p_Docs->col_ptr(i+1); j++)
	  Concept_Vector[Cluster[i]][p_Docs->row_ind(j)]+= p_Docs->val(j);
      else
	Cluster[i] =0;
    }

  if (Alg == SPHERICAL_K_MEANS)
    for (i = 0; i < n_Clusters; i++)
      normalize_vec(Concept_Vector[i], n_Words);
  else
    {
      for (i = 0; i < n_Clusters; i++)
	{
	  CVNormSqr[i] = 0;
	  if (ClusterSize[i] != 0)
	    for (j = 0; j < n_Words; j++)
	      {
		Concept_Vector[i][j] /= ClusterSize[i];
		CVNormSqr[i] += Concept_Vector[i][j]
		  * Concept_Vector[i][j];
	      }
	}
    }
  
  // initial assignment
  for (i = 0; i < n_Docs; i++)
    Inner_Sim[i] = dot_mult(p_Docs, i, Concept_Vector[Cluster[i]]);
  for (i = 0; i < n_Clusters; i++)
    p_Docs->trans_mult(Concept_Vector[i], Sim_Mat[i]);
  
  for (i = 0; i < n_Clusters; i++)
    diff[i] = 0.0;
  
  /* iterate until the stopping criterion is met
     the stopping criterion adopted here is to stop
     if the "change" in objective function after an iteration
     of the algorithm is less than a certain threshold. */
  Result = 0.0;
  n_Iters = 0;
  cout<<"Objective function values are: "<<endl;
  //	pDocs->prepare_trans_mult();
  start_clock = clock();
  do
    {
      assign_cluster(p_Docs);
      compute_cluster_size();
      
      temp_result = Result;
      n_Iters++;
      
      
      // save old concept vectors
      for (i = 0; i < n_Clusters; i++)
	for (j = 0; j < n_Words; j++)
	  old_CV[i][j] = Concept_Vector[i][j];
      
      // reset concept vectors
      for (i = 0; i < n_Clusters; i++)
	for (j = 0; j < n_Words; j++)
	  Concept_Vector[i][j] = 0.0;
      
      // compute new concept vectors
      for (i = 0; i < n_Docs; i++)
	{
	  assert(Cluster[i] >=0 && Cluster[i] < n_Clusters);
	  for (j = p_Docs->col_ptr(i); j < p_Docs->col_ptr(i+1); j++)
	    Concept_Vector[Cluster[i]][p_Docs->row_ind(j)]
					+= p_Docs->val(j);
	}
      
      if (Alg == SPHERICAL_K_MEANS)
	for (i = 0; i < n_Clusters; i++)
	  normalize_vec(Concept_Vector[i], n_Words);
      else
	{
	  for (i = 0; i < n_Clusters; i++)
	    {
	      CVNormSqr[i] = 0;
	      if (ClusterSize[i] != 0)
		for (j = 0; j < n_Words; j++)
		  {
		    Concept_Vector[i][j] /= ClusterSize[i];
		    CVNormSqr[i] += Concept_Vector[i][j]
		      * Concept_Vector[i][j];
		  }
	    }
	}
      
      // compute the difference between old and new concept vectors
      for (i = 0; i < n_Clusters; i++)
	{
	  diff[i] = 0.0;
	  for (j = 0; j < n_Words; j++)
	    diff[i] += (old_CV[i][j] - Concept_Vector[i][j]) *
	      (old_CV[i][j] - Concept_Vector[i][j]);
	  diff[i] = sqrt(diff[i]);
	}
      
      // compute new objective function
      Result = coherence(p_Docs);
      cout<<Result<<endl;
      if (DumpCSize) dump();
      
    } while (//(temp_result > Result || Alg != EUCLIDEAN_K_MEANS) &&
	     (fabs(temp_result - Result) > Epsilon));
 
  finish_clock = clock();
  cout<<"Core kmeans (routine) time: "<<(finish_clock - start_clock)/1e6<<endl;
  float *word_den = new float[n_Clusters];
  for (i = 0; i < n_Clusters; i++)
    {
      word_den[i] = 0.0;
      for (j = 0; j < n_Words; j++)
	if (Concept_Vector[i][j] != 0) word_den[i] += 1.0/n_Words;
    }
  
  cout << "word density of concept vectors: ";
  for (i = 0; i < n_Clusters; i++)
    cout << word_den[i] << " ";
  cout << "\n";
}

void SPKMeans::Iterate(DenseMatrixDouble *p_Docs)
{
  float temp_result;
  int i, j;
  clock_t start_clock, finish_clock;
  
  // normalize document matrix if using spherical k means algorithm
  // else compute norms of document vectors
  if (Alg == SPHERICAL_K_MEANS)
    normalize_mat(p_Docs);
  else
    {
      for (i = 0; i <n_Docs; i++)
	{
	  DocNormSqr[i] = 0;
	  for (j = 0; j<p_Docs->GetNumRow(); j++)
	    DocNormSqr[i] += p_Docs->val(j, i) * p_Docs->val(j, i);
	}
    }
  
  switch (Init_Method)
    {
    case SAMPLE_INIT:
      random_perturb_init(p_Docs);
      //refined_partition(p_Docs, n_Samples);
      break;
    case RANDOM_PERTURB_INIT:
      random_perturb_init(p_Docs);
      break;
    case RANDOM_INIT:
      random_init(p_Docs);
      if (Alg == EUCLIDEAN_K_MEANS)	// should be moved into the concept vector initialization
	{
	  for (i = 0; i < n_Clusters; i++)
	    {
	      CVNormSqr[i] = 0;
	      for (j = 0; j < n_Words; j++)
		CVNormSqr[i] += Concept_Vector[i][j] * Concept_Vector[i][j];
	    }
		}
      break;
    default:
      break;
    }
  
  
  // initial assignment, all documents are assigned to cluster 0
  for (i = 0; i < n_Docs; i++)
    {
      Cluster[i] = 0;
      Inner_Sim[i] = dot_mult(p_Docs, i, Concept_Vector[0]);
    }
  
  for (i = 0; i < n_Clusters; i++)
    p_Docs->trans_mult(Concept_Vector[i], Sim_Mat[i]);
  for (i = 0; i < n_Clusters; i++)
    diff[i] = 0.0;
  
  /* iterate until the stopping criterion is met
     the stopping criterion adopted here is to stop
     if the "change" in objective function after an iteration
     of the algorithm is less than a certain threshold. */
  Result = 0.0;
  n_Iters = 0;

  //cout<<"\n"<<p_Docs->GetNumRow ()<<" "<<n_Words<<p_Docs->val(0,0)<<endl;
  
  
  //	pDocs->prepare_trans_mult();
  start_clock = clock();
  
  do
    {
      assign_cluster(p_Docs);
      compute_cluster_size();
      
      temp_result = Result;
      n_Iters++;
      
      
      // save old concept vectors
      for (i = 0; i < n_Clusters; i++)
	for (j = 0; j < n_Words; j++)
	  old_CV[i][j] = Concept_Vector[i][j];
      
      // reset concept vectors
      for (i = 0; i < n_Clusters; i++)
	for (j = 0; j < n_Words; j++)
	  Concept_Vector[i][j] = 0.0;
      
      // compute new concept vectors
      for (i = 0; i < n_Docs; i++)
	{
	  assert(Cluster[i] >=0 && Cluster[i] < n_Clusters);
	  for (j =0; j< p_Docs->GetNumRow ();  j++)
	    Concept_Vector[Cluster[i]][j]
	      += p_Docs->val(j, i);
	}
      
      if (Alg == SPHERICAL_K_MEANS)
	for (i = 0; i < n_Clusters; i++)
	  normalize_vec(Concept_Vector[i], n_Words);
      else
	{
	  for (i = 0; i < n_Clusters; i++)
	    {
	      CVNormSqr[i] = 0;
	      if (ClusterSize[i] != 0)
		for (j = 0; j < n_Words; j++)
		  {
		    Concept_Vector[i][j] /= ClusterSize[i];
		    CVNormSqr[i] += Concept_Vector[i][j]
		      * Concept_Vector[i][j];
		  }
	    }
	}
      
      // compute the difference between old and new concept vectors
      for (i = 0; i < n_Clusters; i++)
	{
	  diff[i] = 0.0;
	  for (j = 0; j < n_Words; j++)
	    diff[i] += (old_CV[i][j] - Concept_Vector[i][j]) *
	      (old_CV[i][j] - Concept_Vector[i][j]);
	  diff[i] = sqrt(diff[i]);
	}
      
      // compute new objective function
      Result = coherence(p_Docs);
      cout<<Result<<endl;
      if (DumpCSize) dump();
      
    } while (//(temp_result > Result || Alg != EUCLIDEAN_K_MEANS) &&
	     (fabs(temp_result - Result) > Epsilon*Result));
  finish_clock = clock();
  cout<<"Core kmeans (routine) time: "<<(finish_clock - start_clock)/1e6<<endl;
  float *word_den = new float[n_Clusters];
  for (i = 0; i < n_Clusters; i++)
    {
      word_den[i] = 0.0;
      for (j = 0; j < n_Words; j++)
	if (Concept_Vector[i][j] != 0) word_den[i] += 1.0/n_Words;
    }
  
  cout << "word density of concept vectors: ";
  for (i = 0; i < n_Clusters; i++)
    cout << word_den[i] << " ";
  cout << "\n";
}


int SPKMeans::SortCluster(int *cluster_size )
{
  //int *cluster_size = new int[n_Clusters];
	int *cluster_label = new int[n_Clusters];
	int *new_label = new int[n_Clusters];
	int i, j, tmp_size, tmp_label;
	bool flag;

	for (i = 0; i < n_Clusters; i++)
	{
		cluster_size[i] = 0;
		cluster_label[i] = i;
	}
	for (i = 0; i < n_Docs; i++)
		cluster_size[Cluster[i]]++;

	// ugly sorting
	for (i = 0; i < n_Clusters-1; i++)
	{
		flag = true;
		for (j = 0; j < n_Clusters-i-1; j++)
			if (cluster_size[j+1] < cluster_size[j])
			{
				flag = false;
				tmp_size = cluster_size[j+1];
				tmp_label = cluster_label[j+1];
				cluster_size[j+1] = cluster_size[j];
				cluster_label[j+1] = cluster_label[j];
				cluster_size[j] = tmp_size;
				cluster_label[j] = tmp_label;
			}
		if (flag) break;
	}

	// eliminate the empty clusters
	int emptyCluster=0;
	i=0;
	while(cluster_size [i++] ==0)
	  emptyCluster++;
	    
	WordCluster = new int[n_Words];
	int index;
	float max;

	for (j = 0; j < n_Words; j++)
	  {
	    max = Concept_Vector[0][j];
	    index =0;

	    for (i = 1; i < n_Clusters; i++)
	      if (Concept_Vector[i][j] > max) 
		{
		  max = Concept_Vector[i][j];
		  index =i;
		}
	    WordCluster[j]=index;
	  }
	
	// modify Cluster array, also ugly
	for (i = 0; i < n_Clusters; i++)
		new_label[cluster_label[i]] = i;

	for (i = 0; i < n_Docs; i++)
	  Cluster[i] = new_label[Cluster[i]]-emptyCluster;

	for (i = 0; i < n_Words; i++)
	  WordCluster[i] = new_label[WordCluster[i]]-emptyCluster;

	return emptyCluster;
}

void SPKMeans::wordCluster(char *output_matrix, int n_Clusters)
{
  char browserfilepost[128];
  //double max; 
  int j;

  sprintf(browserfilepost, "_wordtoclus.%d", n_Clusters);
  strcat(output_matrix, browserfilepost);
  std::ofstream o_m(output_matrix);
  
  o_m<< n_Words <<endl;

  for (j = 0; j < n_Words; j++)
    o_m<<WordCluster[j]<<endl;
 
  o_m.close();
    
}

void SPKMeans::compute_cluster_size()
{
	int i, k;

	for (i = 0; i < n_Clusters; i++)
	  ClusterSize[i] = 0;
	k=0;
	for (i = 0; i < n_Docs; i++)
	{
	  while (i<empty_Docs_ID[k])
	    {
	      ClusterSize[Cluster[i]]++;
	      i++;
	    }
	  k++;
	}
}

float SPKMeans::coherence(SparseMatrixDouble *p_Docs)
{
	int i;
	float value = 0.0;

	if (Alg == SPHERICAL_K_MEANS)
	{
		if (ObjFunc == NON_WEIGHTED_OBJ)
			for (i = 0; i < n_Docs; i++)
			{
			  Inner_Sim[i] = dot_mult(p_Docs, i, Concept_Vector[Cluster[i]]);
			  
			  value += Inner_Sim[i];
			}
		else
			;	// ...
	}
	else
	{
		for (i = 0; i < n_Docs; i++)
			value += DocNormSqr[i] + CVNormSqr[Cluster[i]] -
				2 * dot_mult(p_Docs, i, Concept_Vector[Cluster[i]]);
	}
	
	return value;
}

float SPKMeans::coherence(DenseMatrixDouble *p_Docs)
{
	int i;
	float value = 0.0;

	if (Alg == SPHERICAL_K_MEANS)
	{
		if (ObjFunc == NON_WEIGHTED_OBJ)
			for (i = 0; i < n_Docs; i++)
			{
				Inner_Sim[i] = dot_mult(p_Docs, i, Concept_Vector[Cluster[i]]);
				value += Inner_Sim[i];
			}
		else
			;	// ...
	}
	else
	{
		for (i = 0; i < n_Docs; i++)
			value += DocNormSqr[i] + CVNormSqr[Cluster[i]] -
				2 * dot_mult(p_Docs, i, Concept_Vector[Cluster[i]]);
	}

	return value;
}

void SPKMeans::assign_cluster(SparseMatrixDouble *p_Docs)
{
	int i, j;
/*	VECTOR_double similarity, temp;

	similarity = new double[n_Docs];
	temp = new double[n_Docs]; */

	float temp_sim;
	int moved = 0;
	int multi = 0;
	int changed = 0;
	int *tc = new int[n_Docs];
	for (i = 0; i < n_Docs; i++) tc[i] = Cluster[i];

	// compute the similarity between concept vectors
	// ...
	// only apply to spherical k means
/*	assert(Alg == SPHERICAL_K_MEANS);
	for (i = 0; i < n_Clusters; i++)
	{
		CV_Sim(i, i) = 1.0;
		for (j = 0; j < i; j++)
			CV_Sim(i, j) = dot_mult(Concept_Vector[i], Concept_Vector[j], n_Words);
	} */

	// estimate the similarity matrix
	for (i = 0; i < n_Clusters; i++)
		for (j = 0; j < n_Docs; j++)
			Sim_Mat[i][j] += diff[i];
	for (i = 0; i < n_Docs; i++)
		Sim_Mat[Cluster[i]][i] = Inner_Sim[i];

	// reassign documents
	for (i = 0; i < n_Docs; i++)
	{
		for (j = 0; j < n_Clusters; j++)
			if (j != Cluster[i])
			{
/*				if (CV_Sim(Cluster[i], j) > 2 * Inner_Sim[i] * Inner_Sim[i]- 1)
				{	// ...
					multi++;
					temp_sim = dot_mult(p_Docs, i, Concept_Vector[j]);
					if (temp_sim > Inner_Sim[i])
					{
						changed++;
						Cluster[i] = j;
						Inner_Sim[i] = temp_sim;
					}
				} */
				if (Sim_Mat[j][i] > Sim_Mat[Cluster[i]][i])
				{
					multi++;
					temp_sim = dot_mult(p_Docs, i, Concept_Vector[j]);
					Sim_Mat[j][i] = temp_sim;
					if (temp_sim > Inner_Sim[i])
					{
						changed++;
						Cluster[i] = j;
						Inner_Sim[i] = temp_sim;
					}
				}
			}
	}

	if(Dump)
	  {
	    cout << multi << " dot product computation\n";
	    cout << changed << " assignment changes\n";
	    for (i = 0; i < n_Docs; i++) if (tc[i] != Cluster[i]) moved++;
	    cout << moved << " documents actually moved\n"<<endl;
	  }
/*
	pDocs->trans_mult(Concept_Vector[0], similarity);
	if (Alg == EUCLIDEAN_K_MEANS)
		for (i = 0; i < n_Docs; i++)	// DocNormSqr is redundant ...
			similarity[i] = DocNormSqr[i] + CVNormSqr[0] - 2 * similarity[i];

	for (i=0; i<n_Docs; i++)
		Cluster[i] = 0;

	for (i=1; i<n_Clusters; i++)
	{
		pDocs->trans_mult(Concept_Vector[i], temp);
		if (Alg == EUCLIDEAN_K_MEANS)
			for (j = 0; j < n_Docs; j++)
				temp[j] = DocNormSqr[j] + CVNormSqr[i]
					- 2 * temp[j];

		for (j = 0; j < n_Docs; j++)
		{
			if (Alg == SPHERICAL_K_MEANS && temp[j] > similarity[j]
				|| Alg == EUCLIDEAN_K_MEANS && temp[j] < similarity[j])
			{
				similarity[j] = temp[j];
				Cluster[j] = i;
			}
		}
	}
*/
}

void SPKMeans::assign_cluster(DenseMatrixDouble *p_Docs)
{
	int i, j;

	float temp_sim;
	int moved = 0;
	int multi = 0;
	int changed = 0;
	int *tc = new int[n_Docs];
	for (i = 0; i < n_Docs; i++) tc[i] = Cluster[i];

	// compute the similarity between concept vectors
	// ...
	// only apply to spherical k means

	
	// estimate the similarity matrix
	for (i = 0; i < n_Clusters; i++)
		for (j = 0; j < n_Docs; j++)
			Sim_Mat[i][j] += diff[i];
	for (i = 0; i < n_Docs; i++)
		Sim_Mat[Cluster[i]][i] = Inner_Sim[i];

	// reassign documents
	for (i = 0; i < n_Docs; i++)
	{
		for (j = 0; j < n_Clusters; j++)
			if (j != Cluster[i])
			{

				if (Sim_Mat[j][i] > Sim_Mat[Cluster[i]][i])
				{
					multi++;
					
					temp_sim = dot_mult(p_Docs, i, Concept_Vector[j]);
					
					Sim_Mat[j][i] = temp_sim;
					if (temp_sim > Inner_Sim[i])
					{
						changed++;
						Cluster[i] = j;
						Inner_Sim[i] = temp_sim;
					}
				}
			}
	}

	cout << multi << " dot product computation\n";
	cout << changed << " assignment changes\n";
	for (i = 0; i < n_Docs; i++) if (tc[i] != Cluster[i]) moved++;
	cout << moved << " documents moved\n";
/*
	pDocs->trans_mult(Concept_Vector[0], similarity);
	if (Alg == EUCLIDEAN_K_MEANS)
		for (i = 0; i < n_Docs; i++)	// DocNormSqr is redundant ...
			similarity[i] = DocNormSqr[i] + CVNormSqr[0] - 2 * similarity[i];

	for (i=0; i<n_Docs; i++)
		Cluster[i] = 0;

	for (i=1; i<n_Clusters; i++)
	{
		pDocs->trans_mult(Concept_Vector[i], temp);
		if (Alg == EUCLIDEAN_K_MEANS)
			for (j = 0; j < n_Docs; j++)
				temp[j] = DocNormSqr[j] + CVNormSqr[i]
					- 2 * temp[j];

		for (j = 0; j < n_Docs; j++)
		{
			if (Alg == SPHERICAL_K_MEANS && temp[j] > similarity[j]
				|| Alg == EUCLIDEAN_K_MEANS && temp[j] < similarity[j])
			{
				similarity[j] = temp[j];
				Cluster[j] = i;
			}
		}
	}
*/
}

// select n_Clusters document vectors as concept vectors
void SPKMeans::random_init(SparseMatrixDouble *p_Docs)
{
	int i, j;
	int *selected;

	selected = new int[n_Clusters];

	// select document vectors
	for (i = 0; i < n_Clusters; i++)
	{
		while (true)
		{
			selected[i] = rand_gen.GetUniformInt(n_Docs);
			j = 0;
			while (j < i && selected[j] != selected[i])
				j++;
			if (j == i)	// this document vector has not been selected
				break;
		}
	}

	// copy document vectors to concept vectors
	for (i = 0; i < n_Clusters; i++)
		for (j = 0; j < n_Words; j++)
			Concept_Vector[i][j] = (*p_Docs)(j, selected[i]);
}

void SPKMeans::random_init(DenseMatrixDouble *p_Docs)
{
	int i, j;
	int *selected;

	selected = new int[n_Clusters];

	// select document vectors
	for (i = 0; i < n_Clusters; i++)
	{
		while (true)
		{
			selected[i] = rand_gen.GetUniformInt(n_Docs);
			j = 0;
			while (j < i && selected[j] != selected[i])
				j++;
			if (j == i)	// this document vector has not been selected
				break;
		}
	}

	// copy document vectors to concept vectors
	for (i = 0; i < n_Clusters; i++)
		for (j = 0; j < n_Words; j++)
			Concept_Vector[i][j] = p_Docs->val(selected[i], j);
}

// nor verified with classic k means ...
void SPKMeans::random_perturb_init(SparseMatrixDouble *p_Docs)
{
	VECTOR_double v, temp_v;
	int i, j;
	float temp_norm;

	v = new float[n_Words];
	temp_v = new float[n_Words];

	for (i = 0; i < n_Words; i++)
		v[i] = 0.0;

	// compute the centroid of document vectors
	for (i = 0; i < n_Docs; i++)
		for (j = p_Docs->col_ptr(i); j < p_Docs->col_ptr(i+1); j++)
			v[p_Docs->row_ind(j)] += p_Docs->val(j);

	if (Alg == SPHERICAL_K_MEANS)
		normalize_vec(v, n_Words);
	else
		for (i = 0; i < n_Words; i++)
			v[i] /= n_Docs;

	// perturb the centroid k times with randomly generated vectors
	for (i = 0; i < n_Clusters; i++)
	{
		//generate a random vector
		for (j = 0; j < n_Words; j++)
			temp_v[j] = rand_gen.GetUniform() - 0.5;
		normalize_vec(temp_v, n_Words);
		temp_norm = Perturb * rand_gen.GetUniform();
		for (j = 0; j < n_Words; j++)
			temp_v[j] *= temp_norm;
		for (j = 0; j < n_Words; j++)
			Concept_Vector[i][j] = v[j]*(1+ temp_v[j]);
		if (Alg == SPHERICAL_K_MEANS)
			normalize_vec(Concept_Vector[i], n_Words);

		// ...
		/* for (j = 0; j < n_Words; j++)
			Concept_Vector[i][j] = docs(j, i); */
	}

/*
#ifdef DEBUG
	cout << "initial concept vectors\n";
	for (i = 0; i < n_Clusters; i++)
	{
		cout << "concept vector " << i << "\n";
		for (j = 0; j < n_Words; j++)
			cout << Concept_Vector[i][j] << " ";
		cout << "\n\n";
	}
#endif */
}

void SPKMeans::random_perturb_init(DenseMatrixDouble *p_Docs)
{
	VECTOR_double v, temp_v;
	int i, j;
	float temp_norm;

	v = new float[n_Words];
	temp_v = new float[n_Words];
	
	for (i = 0; i < n_Words; i++)
		v[i] = 0.0;

	// compute the centroid of document vectors
	for (i = 0; i < p_Docs->GetNumRow (); i++)
		for (j =0; j< p_Docs->GetNumCol (); j++)
			v[i] += p_Docs->val(i, j);

	if (Alg == SPHERICAL_K_MEANS)
		normalize_vec(v, n_Words);
	else
		for (i = 0; i < n_Words; i++)
			v[i] /= n_Docs;

	// perturb the centroid k times with randomly generated vectors
	for (i = 0; i < n_Clusters; i++)
	{
		//generate a random vector
		for (j = 0; j < n_Words; j++)
			temp_v[j] = rand_gen.GetUniform() - 0.5;
		normalize_vec(temp_v, n_Words);
		temp_norm = Perturb * rand_gen.GetUniform();
		for (j = 0; j < n_Words; j++)
			temp_v[j] *= temp_norm;
		for (j = 0; j < n_Words; j++)
			Concept_Vector[i][j] = v[j] + temp_v[j];
		if (Alg == SPHERICAL_K_MEANS)
			normalize_vec(Concept_Vector[i], n_Words);

		// ...
		/* for (j = 0; j < n_Words; j++)
			Concept_Vector[i][j] = docs(j, i); */
	}

/*
#ifdef DEBUG
	cout << "initial concept vectors\n";
	for (i = 0; i < n_Clusters; i++)
	{
		cout << "concept vector " << i << "\n";
		for (j = 0; j < n_Words; j++)
			cout << Concept_Vector[i][j] << " ";
		cout << "\n\n";
	}
#endif */
}

/*
void SPKMeans::refined_partition(SparseMatrixDouble *p_Docs, int j)
{
	int i, ii, iii, best_init;
	float best = 0.0, temp = 0.0;
	int* columns;
	int subsample_size;
	int index;


	VECTOR_double **cm = new (VECTOR_double *)[j];
	for (i = 0; i < j; i++)
	{
		cm[i] = new VECTOR_double[n_Clusters];
		for (ii = 0; ii < n_Clusters; ii++)
			cm[i][ii] = new float[n_Words];
	}

	int nz;
	int* row_ind, *col_ptr;
	float *val;
	SparseMatrixDouble *s, *mcm;

	for (i = 0; i < j; i++)
	{
		subsample_size = int(n_Docs * SampleSize);
		columns = new int[subsample_size];

		nz = 0;

		for ( ii = 0; ii < subsample_size; ii ++)
		{
			columns[ii] = rand_gen.GetUniformInt(n_Docs);
			//create a matrix called s
			nz += p_Docs->col_ptr(columns[ii]+1) - p_Docs->col_ptr(columns[ii]);
		}
		row_ind = new int[nz];
		val = new float[nz];
		col_ptr = new int[subsample_size + 1];

		index = 0;

		for ( ii = 0; ii < subsample_size; ii++)
		{
			col_ptr[ii] = index;
			for ( iii = p_Docs->col_ptr(columns[ii]); iii < p_Docs->col_ptr(columns[ii]+1); iii ++)
			{
				val[index] = p_Docs->val(iii);
				row_ind[index] = p_Docs->row_ind(iii);
				index ++;

			}
		}
		col_ptr[subsample_size] = nz;
		s = new SparseMatrixDouble(n_Words, subsample_size, nz, val, row_ind, col_ptr);

		k_means_mod((*s), cm[i]);
		delete s;
	}
	//create a matrix mcm according to cm

	int *col_ptr1, *row_ind1;
	float *val1;
	nz = 0;
	for ( i = 0; i < n_Clusters; i ++)
	{
		for (iii = 0; iii < j; iii++)
		for (ii = 0; ii < n_Words; ii ++)
		{
			if (cm[iii][i][ii] != 0) nz ++;
		}
	}
	val1 = new float[nz]; col_ptr1 = new int[nz + 1]; row_ind1 = new int[nz];
	index = 0;
	for ( i = 0; i <  j; i ++)
	{
		for (iii = 0; iii < n_Clusters; iii++)
		{
		col_ptr1[i*n_Clusters+iii] = index;
		for (ii = 0; ii < n_Words; ii ++)
		{
			if (cm[i][iii][ii] != 0)
			{
				val1[index] = cm[i][iii][ii];
				row_ind1[index] = ii;
				index ++;
			}
		}
		}
	}
	col_ptr1[n_Clusters * j] = nz;
	mcm = new SparseMatrixDouble(n_Words, n_Clusters * j, nz, val1, row_ind1, col_ptr1);

	SPKMeans K;
	int *cluster = new int[n_Clusters * j];

	for (i = 0; i < j; i ++)
	{
		K.Init(mcm, cluster, n_Clusters, CONCEPT_VECTORS_INIT);
//		for (ii = 0; ii < n_Clusters; ii++)
//			assert(mcm->dim(0) == cm[i][ii].dim());
		K.SetConceptVectors(cm[i]);
		K.Iterate(p_Doc);
		temp = K.GetResult(p_Doc);
		if (best < temp)
		{
			best = temp;
			best_init = i;
		}

	}

	delete mcm;
	for (i = 0; i < n_Clusters; i++)
		for (int k = 0; k < n_Words; k++)
			Concept_Vector[i][k] = cm[best_init][i][k];
	for (i = 0; i < j; i++)
		delete[] cm[i];
	delete[] cm;
}

void SPKMeans::refined_partition(DenseMatrixDouble *p_Docs, int j)
{
	int i, ii, iii, best_init;
	float best = 0.0, temp = 0.0;
	int* columns;
	int subsample_size;
	int index;


	VECTOR_double **cm = new (VECTOR_double *)[j];
	for (i = 0; i < j; i++)
	{
		cm[i] = new VECTOR_double[n_Clusters];
		for (ii = 0; ii < n_Clusters; ii++)
			cm[i][ii] = new float[n_Words];
	}

	int nz;
	int* row_ind, *col_ptr;
	float *val;
	SparseMatrixDouble *s, *mcm;

	for (i = 0; i < j; i++)
	{
		subsample_size = int(n_Docs * SampleSize);
		columns = new int[subsample_size];

		nz = 0;

		for ( ii = 0; ii < subsample_size; ii ++)
		{
			columns[ii] = rand_gen.GetUniformInt(n_Docs);
			//create a matrix called s
			//nz += p_Docs->col_ptr(columns[ii]+1) - p_Docs->col_ptr(columns[ii]);
			nz += p_Doc->GetNumCol();
		}
		row_ind = new int[nz];
		val = new float[nz];
		col_ptr = new int[subsample_size + 1];

		index = 0;

		for ( ii = 0; ii < subsample_size; ii++)
		{
			col_ptr[ii] = index;
			for ( iii = p_Docs->col_ptr(columns[ii]); iii < p_Docs->col_ptr(columns[ii]+1); iii ++)
			{
				val[index] = p_Docs->val(iii);
				row_ind[index] = p_Docs->row_ind(iii);
				index ++;

			}
		}
		col_ptr[subsample_size] = nz;
		s = new SparseMatrixDouble(n_Words, subsample_size, nz, val, row_ind, col_ptr);

		k_means_mod((*s), cm[i]);
		delete s;
	}
	//create a matrix mcm according to cm

	int *col_ptr1, *row_ind1;
	float *val1;
	nz = 0;
	for ( i = 0; i < n_Clusters; i ++)
	{
		for (iii = 0; iii < j; iii++)
		for (ii = 0; ii < n_Words; ii ++)
		{
			if (cm[iii][i][ii] != 0) nz ++;
		}
	}
	val1 = new float[nz]; col_ptr1 = new int[nz + 1]; row_ind1 = new int[nz];
	index = 0;
	for ( i = 0; i <  j; i ++)
	{
		for (iii = 0; iii < n_Clusters; iii++)
		{
		col_ptr1[i*n_Clusters+iii] = index;
		for (ii = 0; ii < n_Words; ii ++)
		{
			if (cm[i][iii][ii] != 0)
			{
				val1[index] = cm[i][iii][ii];
				row_ind1[index] = ii;
				index ++;
			}
		}
		}
	}
	col_ptr1[n_Clusters * j] = nz;
	mcm = new SparseMatrixDouble(n_Words, n_Clusters * j, nz, val1, row_ind1, col_ptr1);

	SPKMeans K;
	int *cluster = new int[n_Clusters * j];

	for (i = 0; i < j; i ++)
	{
		K.Init(mcm, cluster, n_Clusters, CONCEPT_VECTORS_INIT);
//		for (ii = 0; ii < n_Clusters; ii++)
//			assert(mcm->dim(0) == cm[i][ii].dim());
		K.SetConceptVectors(cm[i]);
		K.Iterate(p_Doc);
		temp = K.GetResult(p_Doc);
		if (best < temp)
		{
			best = temp;
			best_init = i;
		}

	}

	delete mcm;
	for (i = 0; i < n_Clusters; i++)
		for (int k = 0; k < n_Words; k++)
			Concept_Vector[i][k] = cm[best_init][i][k];
	for (i = 0; i < j; i++)
		delete[] cm[i];
	delete[] cm;
}

void Assign_Cluster(SparseMatrixDouble &docs, VECTOR_double concept_vectors[], int k,
			int cluster[])
{
	int i, j;
	VECTOR_double similarity, temp;

	similarity = new float[docs.GetNumCol()];
	temp = new float[docs.GetNumCol()];

	docs.trans_mult(concept_vectors[0], similarity);
	for (i=0; i<docs.GetNumCol(); i++)
		cluster[i] = 0;

	for (i=1; i<k; i++)
	{
		docs.trans_mult(concept_vectors[i], temp);
		for (j = 0; j < docs.GetNumCol(); j++)
		{
			if (temp[j] > similarity[j])
			{
				similarity[j] = temp[j];
				cluster[j] = i;
			}
		}
	}
}

VECTOR_double SPKMeans::farthest_data(SparseMatrixDouble &docs, VECTOR_double concept_vector[])
{
	int i;
	float min_similarity = 1.0, temp;
	int farthest = -1;
	VECTOR_double result = new float[docs.GetNumRow()];
	int *cluster = new int[docs.GetNumCol()];

	for (i = 0; i < docs.GetNumRow(); i++)
		result[i] = 0.0;
	Assign_Cluster(docs, concept_vector, n_Clusters, cluster);
	for (i = 0; i < docs.GetNumCol(); i++)
	{
		temp = dot_mult(&docs, i, concept_vector[cluster[i]]);
		if (temp < min_similarity)
		{
			min_similarity = temp;
			farthest = i;
		}
	}

	for (i = docs.col_ptr(farthest); i < docs.col_ptr(farthest+1); i++)
	{
		result[docs.row_ind(i)] = docs.val(i);
	}

	return result;
}

void SPKMeans::k_means_mod(SparseMatrixDouble &docs, VECTOR_double result_concept_vectors[])
{
	assert (docs.GetNumRow() == n_Words);

	int i, j;
	int* cluster_size, *cluster;
	cluster_size = new int[n_Clusters];
	cluster = new int[docs.GetNumCol()];
	int redo = 0; // boolean
	VECTOR_double *temp_concept_vectors = new VECTOR_double[n_Clusters];
	for (i = 0; i < n_Clusters; i++)
		temp_concept_vectors[i] = new float[docs.GetNumRow()];
	for ( i = 0; i < n_Clusters; i ++)
		for (j = 0; j < docs.GetNumRow(); j++)
			temp_concept_vectors[i][j] = Concept_Vector[i][j];
	SPKMeans K;

	do {
	redo = 0;
	K.Init(&docs, cluster, n_Clusters, CONCEPT_VECTORS_INIT);
	K.SetConceptVectors(temp_concept_vectors);
	K.Iterate();

	for ( i = 0; i < n_Clusters; i ++ ) cluster_size[i] = 0;
	for ( i = 0; i < docs.GetNumCol(); i++)
		cluster_size[cluster[i]]++;

	for ( i = 0; i < n_Clusters; i ++ )
	{
		if (cluster_size[i] == 0)
		{
			temp_concept_vectors[i] = farthest_data(docs, temp_concept_vectors);
			redo = 1;
		}
	}
	} while (redo == 1);

	for ( i = 0; i < n_Clusters; i++ )
		for (j = 0; j < docs.GetNumRow(); j++)
			result_concept_vectors[i][j] = temp_concept_vectors[i][j];
}

*/
