/*	SPKMeans.h
	
	Data structures used in spherical k-means algorithm
*/

#if !defined(_SPKMEANS_H_)
#define _SPKMEANS_H_

#include "DenseMatrixDouble.h"
//#include "SparseMatrixDouble.h"
#include "RandomGenerator.h"
#include <iostream.h>

// initialization methods of concept vectors

#define EUCLIDEAN_K_MEANS			1
#define SPHERICAL_K_MEANS		2

#define RANDOM_PERTURB_INIT		1
#define SAMPLE_INIT			2
#define CONCEPT_VECTORS_INIT		3
#define RANDOM_INIT			4
#define GRAPH_PART                      5
#define FILE_INIT                       6

#define DEFAULT_EPSILON			0.001
#define DEFAULT_PERTURB			0.1
#define DEFAULT_NUM_SAMPLE		4
#define DEFAULT_SAMPLE_SIZE		0.1

#define NON_WEIGHTED_OBJ		1
#define WEIGHTED_OBJ			2

#define CD_QR                           50
#define CON_DEC                         51
#define QR_DEC                          52

class SPKMeans
{
public:
	// document matrix will be normalized if using spherical k-means algorithm
	// clustering result are returned in cluster[]
	// init_method selects the initialization method of concept vectors
	// epsilon is the stopping threshold of the k-means algorithm
	SPKMeans(SparseMatrixDouble *p_docs, int cluster[], int num_clusters,
		int init_method, float epsilon = DEFAULT_EPSILON);
	SPKMeans(DenseMatrixDouble *p_docs, int cluster[], int num_clusters,
		int init_method, float epsilon = DEFAULT_EPSILON);
	SPKMeans();
	~SPKMeans();
	bool spkmeans_worked; 
	float ** getQTA();
	float ** getCTA();
	void checkQTA(SparseMatrixDouble *p_Docs);
	void checkCTA(SparseMatrixDouble *p_Docs);

	// ugly, add some assertions
	void SetAlgorithm(int alg) { Alg = alg; }
	void setNodump(bool d){ Dump =d ;};
	void SetEpsilon(float epsilon) { Epsilon = epsilon; }
	void SetPerturb(float perturb) { Perturb = perturb; }
	void SetInitMethod(int init_method) { Init_Method = init_method; }
	void SetConceptVectors(VECTOR_double *vec) { Concept_Vector = vec; }
	void SetNumSamples(int num_samples) { n_Samples = num_samples; }
	void SetSampleSize(float sample_size) { SampleSize = sample_size; }
	void SetEncodeScheme(int encode_scheme) { EncodeScheme = encode_scheme; }
	void SetObjFunc(int obj_func) { ObjFunc = obj_func; }
	void SetDump(bool dump_csize) { DumpCSize = dump_csize; }

	int GetNumIterations() { return n_Iters; }
	float GetResult() { return Result; }
	void Iterate(SparseMatrixDouble *p_Docs, char * filename);
	void Init(SparseMatrixDouble *p_docs, int cluster[], int num_clusters,
			int init_method, float epsilon = DEFAULT_EPSILON); 
	void Init(DenseMatrixDouble *p_docs, int cluster[], int num_clusters,
			int init_method, float epsilon = DEFAULT_EPSILON); 
	void Iterate(SparseMatrixDouble *p_Docs, char* file1, char* file2);			// iterative computation of concept vectors
	void Iterate(DenseMatrixDouble *p_Docs);	
	void wordCluster(char *output_matrix, int n_Clusters);
	int SortCluster(int * cluster_size);		// sort clusters according to their sizes in non-decreasing order


	void Concept_Dec(SparseMatrixDouble *p_Docs,  char * filenamePrefix, int meth);
	//void ReturnConVec(VECTOR_double *ConVec);

	void SetEmptyDocs(int n_E_Docs, int *e_D_ID) 
	  {
	    n_Empty_Docs =n_E_Docs ;
	    
	    empty_Docs_ID = e_D_ID;
	   
	  };
	void outputClusterSize() 
	  { for (int i=0; i< n_Clusters; i++)
	    cout << "cluster " << i<< " : " << ClusterSize[i] << "\n";};
protected:
	int Alg;			// algorithm: classic or spherical
	bool Dump;
	int n_Empty_Docs;
	int *empty_Docs_ID;
	//SparseMatrixDouble *pDocs;	// pointer to document matrix
	//DenseMatrixDouble *DpDocs;	
	int *Cluster;			// which cluster a document belongs to
	int *WordCluster;
	VECTOR_double *Concept_Vector;	// concept vectors
	VECTOR_double *qTransA_Vector;
	VECTOR_double *cTransA_Vector;
	float **old_CV;
	float *diff;
	int *ClusterSize;
	float *DocNormSqr;		// square of 2-norm of document vectors
	float *CVNormSqr;		// square of 2-norm of concept vectors
	
	RandomGenerator_MT19937 rand_gen;
	int n_Words, n_Docs;	// numbers of words and documents of document matrix
	int Init_Method;		// concept vector initialization method
	int n_Clusters;			// number of clusters, equal to number of concept vectors
	int n_Iters;			// number of iterations
	int n_Samples;
	float Epsilon;			// stopping criteria of k-means algorithm
	float Perturb;			// maximum norm of the perturb vectors
	float SampleSize;
	int EncodeScheme;
	int ObjFunc;

	float *Inner_Sim;		// the similarity between a document and the cluster 
					// centroid (the concept vector)
	//float *CV_Sim;			// similarity between concept vectors
	float **Sim_Mat;		// similarity matrix

	float Result;			// final objective function value


	int remove_Empty_Clusters(SparseMatrixDouble *p_Docs);


	float coherence(SparseMatrixDouble *p_Docs);		// the objective function, return the sum of dot products
					// (cosine similarity) of document vectors with their respective concept vectors
	float coherence(DenseMatrixDouble *p_Docs);
	void assign_cluster(SparseMatrixDouble *p_Docs);		// assign document vectors to clusters according
					// to their distance to concept_vectors.
	void assign_cluster(DenseMatrixDouble *p_Docs);
	void compute_cluster_size();
	void random_perturb_init(SparseMatrixDouble *p_Docs);	// initialize concept vectors by perturbing the centroid
					// of document vectors, the result concept vectors are normalized
	void random_perturb_init(DenseMatrixDouble *p_Docs);
	void random_init(SparseMatrixDouble *p_Docs);		// totally random initialization
	void random_init(DenseMatrixDouble *p_Docs);
	// methods used in sampling initialization
	//void refined_partition(SparseMatrixDouble *p_Docs, int j); 
	//void refined_partition(DenseMatrixDouble *p_Docs, int j); 
	//VECTOR_double farthest_data(SparseMatrixDouble &docs, VECTOR_double concept_vector[]);
	//void k_means_mod(SparseMatrixDouble &docs, VECTOR_double result_concept_vectors[]);
	int read_cate(char *fname, int * label);
	void generate_Confusion_Matrix (int *label, int n_Class);
	bool DumpCSize;
	void dump();
};

#endif // !defined(_SPKMEANS_H_)
