/*	mat_vec.cc

	Implementation of the Sparse_Mat_double class

	Copyright(c) 2000 Yuqiang Guan 
*/

#include <iostream.h>
#include <math.h>
#include "mat_vec.h"

void ranksa_c( float arrin[], int indx[], int *n )                              
{                                                                               
   int ir;                                                                      
   int l;                                                                       
                                                                                
   /*                                                                           
    * initialize the index aray with consecutive integers.                      
    */                                                                             
    for (l = 0; l < (*n); l++) indx[l] = l;                                         
    if ((*n) < 2) return;                                                        
   /*                                                                           
    * The index L will be decremented from its initial value down to 1          
    * during the "hiring" (heap creation) phase. Once it reaches 1, the         
    * index IR will be decremented from its initial value down to 1             
    * during the "retirement-and-promotion" (heap selection) phase.             
    */                                                                          
   l = (*n) / 2;  
   
   ir = (*n) - 1;                                                               
   while (1) {                                                                  
      int   i, j;                                                               
      int   indxt;                                                              
      float q;                                                                  
                                                                                
      if (l > 0) {                      /* Still in hiring phase */             
         indxt = indx[--l];                                                              
	 q = arrin[indxt];                                                            
	 } 
	 else {                          /* In retirement-and-promotion phase */ 
         indxt = indx[ir];              /* Clear a space at end of array */     
         q = arrin[indxt];                                                      
         indx[ir--] = indx[0];          /* Retire the top of the heap into it   */                                                                              
         if (ir == 0) {                 /* Done with the last promotion */      
            indx[0] = indxt;            /* The least competent worker of all */ 
            return;              
	    
        }                                                                      
      }                                                                         
      /*                                                                        
       * Whether we are in the hiring phase or promotion phase, we here         
       * set up to sift down element q to its proper level.                     
       */                                                                       
      i = l;                                                                    
      j = l + l + 1;                                                            
      while (j <= ir) {                 /* * Compare to the better underling */                                                                    
         if (j < ir && arrin[indx[j]] < arrin[indx[j+1]]) 
	 	j++;                  
         if (q < arrin[indx[j]]) {      /* Demote q */                          
            indx[i] = indx[j];                                                  
            i = j;                                                              
            j = j + j + 1;        
	    
	    } else {                       /* This is q's level */                 
            j = ir + 1;                 /* Set j to terminate the sift-down */  
         }                                                                      
      }                                                                         
      indx[i] = indxt;                  /* Put q into its slot */               
   }                                                                            
}    	


/* The procedure dqrbasis outputs an orthogonal basis spanned by the rows
   of matrix a (using the QR Factorization of a ).  */
void dqrbasis( int m, int n, float **a, float **q , float *work)
{
  int i,j;

  for(i = 0; i < m;i++){
    dmatvec( i, n, q, a[i], work );
    dmatvecat( i, n, q, work, q[i] );
    for(j = 0; j < n;j++)
      q[i][j] = a[i][j] - q[i][j];
    dvec_l2normalize( n, q[i] );
  }
}

/* Does y = A * x, A is mxn, and y & x are mx1 and nx1 vectors
   respectively  */
void dmatvec(int m, int n, float **a, float *x, float *y)
{
  float yi;
  int i,j;

  for(i = 0;i < m;i++){
    yi = 0.0;
    for(j = 0;j < n;j++){
      yi += a[i][j] * x[j];
    }
    y[i] = yi;
  }
}

/* Does y = A' * x, A is mxn, and y & x are nx1 and mx1 vectors
   respectively  */
void dmatvecat(int m, int n, float **a, float *x, float *y)
{
  float yi;
  int i,j;

  for(i = 0;i < n;i++){
    yi = 0.0;
    for(j = 0;j < m;j++){
      yi += a[j][i] * x[j];
    }
    y[i] = yi;
  }
}

/* The function dvec_l2normsq computes the square of the
   Euclidean length (2-norm) of the double precision vector v */
float dvec_l2normsq( int dim, float *v )
{
  float length,tmp;
  int i;

  length = 0.0;
  for(i = 0;i < dim;i++){
    tmp = *v++;
    length += tmp*tmp;
  }
  return(length);
}

/* The function dvec_l2normalize normalizes the double precision
   vector v to have 2-norm equal to 1 */
void dvec_l2normalize( int dim, float *v )
{
  float nrm;

  nrm = sqrt(dvec_l2normsq( dim, v ));
  if( nrm != 0 ) dvec_scale( 1.0/nrm, dim, v );
}

void dvec_scale( float alpha, int n, float *v )
{
  int i;

  for(i = 0;i < n;i++){
    *v++ = *v * alpha;
  }
}

void normalize_mat(SparseMatrixDouble *mat)
{
	int i, j;
	float norm;

	for (i = 0; i < mat->GetNumCol(); i++)
	{
		norm = 0.0;
		for (j = mat->col_ptr(i); j < mat->col_ptr(i+1); j++)
		norm += mat->val(j) * mat->val(j);
//              assert(norm > 0);
		/*		if (norm == 0)
		{
			cerr << "column " << i << " has 0 norm\n";
			}*/
		if(norm!=0)
		{
		  norm = sqrt(norm);
		  for (j = mat->col_ptr(i); j < mat->col_ptr(i+1); j++)
		    mat->val(j) /= norm;
	        }
	}
}

void normalize_mat(DenseMatrixDouble *mat)
{
  int i, j;
  float norm;

  for ( i=0; i < mat->GetNumCol(); i++)
    {
      norm = 0.0;
      for (j=0; j < mat->GetNumRow(); j++)
	norm+=mat->val(j, i)*mat->val(j, i);
    
      if(norm!=0)
	{
	  norm = sqrt(norm);
	  for (j=0; j<mat->GetNumRow() ; j++)
	    mat->val(j, i) /= norm;
	}
    }
}

void encode_mat(SparseMatrixDouble *mat, int scheme)
{
	int i;
	int *n_doc_per_word;	// number of documents which contains a word
	float *global_term;
	
	switch (scheme)
	{
	case NORM_TERM_FREQ:
		normalize_mat(mat);
		break;
	case NORM_TERM_FREQ_INV_DOC_FREQ:
		n_doc_per_word = new int[mat->GetNumRow()];
		global_term = new float[mat->GetNumRow()];

		for (i = 0; i < mat->GetNumRow(); i++);
			n_doc_per_word[i] = 0;
		for (i = 0; i < mat->GetNumNonzeros(); i++)
			if (mat->row_ind(i) != 0)
				n_doc_per_word[mat->row_ind(i)]++;
		for (i = 0; i < mat->GetNumRow(); i++);
		{
			if (n_doc_per_word[i] == 0)
				global_term[i] = 0;
			else
				global_term[i]
					= log((float)mat->GetNumCol()/n_doc_per_word[i]);
		}
		for (i = 0; i < mat->GetNumNonzeros(); i++)
			mat->val(i) *= mat->val(i) * global_term[mat->row_ind(i)];
		normalize_mat(mat);

		break;
	default:
		break;
	}
}
void encode_mat(DenseMatrixDouble *mat, int scheme)
{
	int i, j;
	int *n_doc_per_word;	// number of documents which contains a word
	float *global_term;
	
	switch (scheme)
	{
	case NORM_TERM_FREQ:
		normalize_mat(mat);
		break;
	case NORM_TERM_FREQ_INV_DOC_FREQ:
		n_doc_per_word = new int[mat->GetNumRow()];
		global_term = new float[mat->GetNumRow()];

		for (i = 0; i < mat->GetNumRow(); i++);
			n_doc_per_word[i] = 0;
		for (i = 0; i < mat->GetNumRow(); i++)
		  for(j=0; j<mat->GetNumCol(); j++)
			if (mat->val(i, j) != 0)
				n_doc_per_word[i]++;
		for (i = 0; i < mat->GetNumRow(); i++);
		{
			if (n_doc_per_word[i] == 0)
				global_term[i] = 0;
			else
				global_term[i]
					= log((float)mat->GetNumCol()/n_doc_per_word[i]);
		}
		for (i = 0; i < mat->GetNumRow(); i++)
		  for(j=0; j<mat->GetNumCol(); j++)
	 	      mat->val(i, j) *= mat->val(i, j) * global_term[i];
		normalize_mat(mat);

		break;
	default:
		break;
	}
}


// normalize the norm of vector v to 1 if it was positive
/*void normalize_vec(float *vec, int size)
{
	float norm;
	int i;

	norm = 0.0;
	for (i = 0; i < size; i++)
	norm += vec[i] * vec[i];

//      assert(norm > 0);
	if (norm == 0) return;
	norm = sqrt(norm);

	for (i = 0; i < size; i++)
	vec[i] /= norm;
}*/

void normalize_vec(float vec[], int n)
{
        float norm;
        int i;

        norm = 0.0;
        for (i = 0; i < n; i++)
        norm += vec[i] * vec[i];

//      assert(norm > 0);
        if (norm == 0) return;
        norm = sqrt(norm);

        for (i = 0; i < n; i++)
        vec[i] /= norm;
}

// compute the dot product of the col'th document vector of docs and concept_vector
float dot_mult(SparseMatrixDouble *mat, int col, float *vec)
{
	float result = 0.0;
	int j;

	for (j = mat->col_ptr(col); j < mat->col_ptr(col+1); j++)
		result += mat->val(j) * vec[mat->row_ind(j)];

	return result;
}

float dot_mult(DenseMatrixDouble *mat, int col, float *vec)
{
  float result = 0.0;
  int j;
  
  for(j=0; j<mat->GetNumRow(); j++)
      result+=mat->val(j, col)*vec[j];
  return result;
}

float dot_mult(float *v1, float *v2, int n)
{
	float result = 0.0;
	int j;

	for (j = 0; j < n; j++)
		result += v1[j] * v2[j];

	return result;
}
